from privacyscanner.scanmodules.testsslsh.base import TestsslshScanModuleBase
from privacyscanner.scanmodules.testsslsh.scanner import Parameter


class TestsslshHttpsScanModule(TestsslshScanModuleBase):
    name = 'testssl_https'
    required_keys = ['final_url', 'https', 'testssl_https']
    target_type = 'https'
    target_parameters = []

    def _get_host(self, result):
        return result['final_url']


class TestsslshMailScanModule(TestsslshScanModuleBase):
    name = 'testssl_mail'
    required_keys = ['mail', 'testssl_mail']
    target_type = 'mail'
    target_parameters = [Parameter.STARTTLS, 'smtp']

    def _get_host(self, result):
        return result['mail']['domain'] + ':25'



QUEUE_DB_DSN = 'dbname=privacyscore user=privacyscore password=privacyscore host=localhost'
MAX_EXECUTION_TIMES = {None: 300}
SCAN_MODULE_OPTIONS = {}
SCAN_MODULES = ['privacyscanner.scanmodules.chromedevtools.ChromeDevtoolsScanModule',
                'privacyscanner.scanmodules.dns.DNSScanModule',
                'privacyscanner.scanmodules.mail.MailScanModule',
                'privacyscanner.scanmodules.serverleaks.ServerleaksScanModule',
                'privacyscanner.scanmodules.testsslsh.TestsslshHttpsScanModule',
                'privacyscanner.scanmodules.testsslsh.TestsslshMailScanModule',]
NUM_WORKERS = 2
MAX_EXECUTIONS = 100
RAVEN_DSN = None
MAX_TRIES = 3
STORAGE_PATH = '~/.local/share/privacyscanner'



from .certificate import CertificateExtractor
from .cookies import CookiesExtractor
from .cookiestats import CookieStatsExtractor
from .failedrequests import FailedRequestsExtractor
from .finalurl import FinalUrlExtractor
from .googleanalytics import GoogleAnalyticsExtractor
from .insecurecontent import InsecureContentExtractor
from .javascriptlibs import JavaScriptLibsExtractor
from .redirectchain import RedirectChainExtractor
from .requests import RequestsExtractor
from .securityheaders import SecurityHeadersExtractor
from .thirdparties import ThirdPartyExtractor
from .tlsdetails import TLSDetailsExtractor
from .trackerdetect import TrackerDetectExtractor
from .screenshot import ScreenshotExtractor
from .imprint import ImprintExtractor



from privacyscanner.scanmodules.chromedevtools.extractors.base import Extractor


class FinalUrlExtractor(Extractor):
    def extract_information(self):
        self.result['final_url'] = self.page.final_response['url']



import io
import re
import tarfile
from pathlib import Path

from privacyscanner.exceptions import RescheduleLater
from privacyscanner.scanmodules import ScanModule
from privacyscanner.scanmodules.testsslsh.scanner import TestsslshScanner, Parameter, TestsslshFailed, \
    TestsslshFailedPartially
from privacyscanner.utils import set_default_options, download_file
from privacyscanner.utils.tls import get_cipher_info


DOWNLOAD_URL = 'https://github.com/drwetter/testssl.sh/archive/3.0rc3.tar.gz'
DOWNLOAD_HASH = '45f8aed24ad749175608a29c50566240a8a1b8ebcb32531d7bf6231ec269f4a5'

_TESTSSL_PROTOCOL_NEGOTIATED_REGEXP = re.compile('^Default protocol (TLS1.[0-3]|SSLv[23])$')
_TESTSSL_SUITE_NAME = re.compile('^([A-Za-z0-9-]+)(,| |$)')
_TESTSSL_DH = re.compile('(\d+) bit (DH|ECDH)')
_TESTSSL_ECDH_CURVE = re.compile(r'\(([^)]+)\)')


class IncompleteStage(Exception):
    def __init__(self, partial_result):
        self.partial_result = partial_result
        super().__init__()


class TestsslshScanModuleBase(ScanModule):
    name: str
    required_keys: list
    target_type: str
    target_parameters: list
    dependencies = ['chromedevtools', 'dns']

    def __init__(self, options):
        set_default_options(options, {
            'install_base_dir': options['storage_path'] / 'testsslsh',
            'download_url': DOWNLOAD_URL,
            'download_hash': DOWNLOAD_HASH,
            'stages': ['basic', 'vulns', 'vulns_ids'],
        })

        for stage in options['stages']:
            if not hasattr(self, '_scan_stage_' + stage):
                raise ValueError('Invalid stage: `{}`.'.format(stage))

        super().__init__(options)
        self._install_dir = self.options['install_base_dir'] / self.options['download_hash']

    def scan_site(self, result, meta):
        stages = self.options['stages']
        testssl_key = 'testssl_' + self.target_type
        if testssl_key not in result:
            result[testssl_key] = {
                'current_stage': stages[0],
                'stages': {}
            }
        result.mark_dirty(testssl_key)
        testssl = result[testssl_key]
        stage_key = testssl['current_stage']

        self.logger.info('Current stage: %s', stage_key)
        if stage_key not in stages:
            self.logger.error('Stage `%s` is not available', stage_key)
            return

        if stage_key not in testssl['stages']:
            testssl['stages'][stage_key] = {'status': 'open'}
        stage_dict = testssl['stages'][stage_key]

        scan_result = None
        try:
            stage_method = getattr(self, '_scan_stage_' + stage_key)
            host = self._get_host(result)
            scan_result = stage_method(host, self.target_parameters)
        except IncompleteStage as e:
            self.logger.info('testssl.sh result is incomplete.')
            scan_result = e.partial_result
            stage_dict['status'] = 'incomplete'
        except TestsslshFailed as e:
            self.logger.error('testssl.sh failed with exit code %s: %s',
                              e.exit_code, e)
            stage_dict['status'] = 'failed'
            stage_dict['error_code'] = e.exit_code
            stage_dict['error_message'] = str(e)
        except Exception as e:
            stage_dict['status'] = 'failed_with_bug'
            stage_dict['error_code'] = -2000
            stage_dict['error_message'] = '{}: {}'.format(e.__class__.__name__, e)
            self.logger.exception('Programming error in testsslsh')
            raise
        else:
            self.logger.info('testssl.sh result is complete.')
            stage_dict['status'] = 'complete'

        if scan_result:
            target_result = result[self.target_type]
            for key, value in scan_result.items():
                if key in target_result:
                    continue
                target_result[key] = value
            result.mark_dirty(self.target_type)

        has_failed = stage_dict['status'] != 'complete'
        if has_failed:
            del testssl['current_stage']
            return

        try:
            next_stage = stages[stages.index(stage_key) + 1]
        except IndexError:
            del testssl['current_stage']
            self.logger.info('%s was the final stage.', stage_key)
            return
        self.logger.info('Next stage: %s', next_stage)
        testssl['current_stage'] = next_stage
        raise RescheduleLater(10)

    def _scan_stage_basic(self, target, extra_parameters):
        """Stage 0 scan: Contains the most relevant checks.

        These include:
        - Protocols (SSLv3, TLSv1.0 - TLSv1.3)
        - Standard ciphers, including weak ciphers (RC4, ...)
        - Forward Secrecy support

        """
        scanner = TestsslshScanner(self._install_dir)
        scanner.add_parameters(Parameter.PROTOCOLS,
                               Parameter.STANDARD_CIPHERS,
                               Parameter.CHECK_FORWARD_SECRECY,
                               Parameter.SERVER_DEFAULTS,
                               Parameter.SERVER_PREFERENCE,
                               Parameter.PHONE_OUT,
                               *extra_parameters)
        try:
            scan_result = scanner.scan(target)
            incomplete = False
        except TestsslshFailedPartially as e:
            scan_result = e.partial_result
            incomplete = True
        findings = ScanResultFindings(scan_result, self.logger)

        forward_secrecy = {
            'available': None,
            'ciphers': None
        }
        protocols = {
            'SSL 2': None,
            'SSL 3': None,
            'TLS 1.0': None,
            'TLS 1.2': None,
            'TLS 1.3': None,
        }
        ciphers = {
            '128Bit': None,
            '3DES': None,
            'DES_and_64Bit': None,
            'EXPORT': None,
            'HIGH': None,
            'NULL': None,
            'STRONG': None,
            'aNULL': None,
        }
        ocsp = {
            'valid': None,
            'stapling': None,
            'must_staple': None,
        }
        resumption = {
            'session_id': None,
            'session_ticket': None,
        }
        certificate_transparency = {
            'valid': None,
            'has_extension': None
        }
        cipher_order = {}
        tls_result = {
            'forward_secrecy': forward_secrecy,
            'certificate_transparency': certificate_transparency,
            'prefer_server_ciphers': None,
            'cipher_order': cipher_order,
            'ocsp': ocsp,
            'supported_protocols': protocols,
            'supported_cipherlists': ciphers,
            'resumption': resumption,
        }

        pfs = findings.get('PFS', ('offered', 'not offered'))
        if pfs:
            forward_secrecy['available'] = pfs == 'offered'

        pfs_ciphers = findings.get('PFS_ciphers')
        if pfs_ciphers:
            forward_secrecy['ciphers'] = pfs_ciphers.split()

        pfs_curves = findings.get('PFS_ECDHE_curves')
        if pfs_curves:
            forward_secrecy['curves'] = pfs_curves.split()

        protocol_name_map = {
            'SSLv2': 'SSL 2',
            'SSLv3': 'SSL 3',
            'TLS1': 'TLS 1.0',
            'TLS1_1': 'TLS 1.1',
            'TLS1_2': 'TLS 1.2',
            'TLS1_3': 'TLS 1.3'
        }
        for source_key, target_key in protocol_name_map.items():
            protocol = findings.get(source_key, (
                'offered',
                'not offered',
                'is not offered', # Why, testssl.sh, WHY?
                'not offered and downgraded to a weaker protocol'
            ))
            if protocol is None:
                continue
            protocols[target_key] = protocol == 'offered'

            # Why, testssl.sh, WHY?
            order_key = source_key.replace('TLS', 'TLSv')
            cipherorder_proto = findings.get('cipherorder_' + order_key)
            if cipherorder_proto:
                cipher_order[target_key] = cipherorder_proto.split()

        # We need protocol to be defined when looking at the cipher that has
        # been negotiated. Normally, protocol will be overriden in this
        # block.
        protocol = None
        protocol_negotiated = findings.get('protocol_negotiated')
        if protocol_negotiated:
            if 'Default protocol' in protocol_negotiated:
                match = _TESTSSL_PROTOCOL_NEGOTIATED_REGEXP.match(protocol_negotiated)
                if match:
                    protocol = protocol_name_map[match.group(1).replace('.', '_')]
                    tls_result['protocol'] = protocol

        cipher_negotiated = findings.get('cipher_negotiated')
        if cipher_negotiated:
            cipher = {
                'cipher': None,
                'key_exchange': None,
                'key_exchange_group': None,
                'mac': None
            }
            match_dh = _TESTSSL_DH.search(cipher_negotiated)
            match_curve = _TESTSSL_ECDH_CURVE.search(cipher_negotiated)
            match_suite = _TESTSSL_SUITE_NAME.search(cipher_negotiated)
            curve = None
            if match_suite:
                bits = 0
                if match_dh:
                    bits = int(match_dh.group(1))
                    dh = match_dh.group(2)
                    if match_curve and dh == 'ECDH':
                        curve = match_curve.group(1)
                cipher_tuple = (match_suite.group(1), protocol, bits)
                cipher.update(get_cipher_info(cipher_tuple))
                cipher['key_exchange_group'] = curve
            tls_result.update(cipher)

        cipher_map = {'DES_and_64Bit': 'cipherlist_DES+64Bit'}
        for key in ciphers:
            source_key = cipher_map.get(key, 'cipherlist_' + key)
            cipherlist = findings.get(source_key, ('offered', 'not offered'))
            if cipherlist is None:
                continue
            ciphers[key] = cipherlist == 'offered'

        cipher_order = findings.get('cipher_order',
                                    ('server', 'NOT cipher order configured'))
        if cipher_order:
            tls_result['prefer_server_ciphers'] = cipher_order == 'server'

        ocsp_stapling = findings.get('OCSP_stapling')
        if ocsp_stapling:
            # Note: We have no assertions here, because testssl.sh uses a
            # dynamic string here. Currently, only 'offered' is the true-case.
            ocsp['stapling'] = ocsp_stapling == 'offered'

        ocsp_revoked = findings.get('cert_ocspRevoked')
        if ocsp_revoked:
            # Again, no _assert_findings. Options are either 'revoked',
            # 'not revoked', or the responder's response.
            if ocsp_revoked == 'revoked':
                ocsp['valid'] = False
            elif ocsp_revoked == 'not revoked':
                ocsp['valid'] = True
            else:
                ocsp['response'] = ocsp_revoked

        ocsp_must_staple = findings.get('cert_mustStapleExtension', (
            '--', 'supported', 'extension detected but no OCSP stapling provided'
        ))
        if ocsp_must_staple:
            ocsp['must_staple'] = ocsp_must_staple != '--'

        session_id_resumption = findings.get('sessionresumption_ID', (
            'No Session ID, no resumption',
            'supported',
            'not supported',
            "check couldn't be performed because of client authentication",
            'check failed, pls report'
        ))
        if session_id_resumption:
            if session_id_resumption == 'supported':
                resumption['session_id'] = True
            elif session_id_resumption in ('No Session ID, no resumption', 'not supported'):
                resumption['session_id'] = False

        session_ticket_support = findings.get('sessionresumption_ticket', (
            'supported',
            'not supported',
            "check couldn't be performed because of client authentication",
            'check failed, pls report'
        ))
        if session_ticket_support:
            if session_ticket_support == 'supported':
                resumption['session_ticket'] = True
            elif session_ticket_support == 'not supported':
                resumption['session_ticket'] = False

        finding_ct = findings.get('certificate_transparency')
        if finding_ct:
            if finding_ct == 'yes (certificate extension)':
                certificate_transparency['has_extension'] = True

        if incomplete:
            raise IncompleteStage(tls_result)
        return tls_result

    def _scan_stage_vulns(self, target, extra_parameters):
        """Stage 1 scan: Contains vulnerabilities which are IDS-proof"""
        scanner = TestsslshScanner(self._install_dir)
        scanner.add_parameters(Parameter.VULN_RENEGOTIATION,
                               Parameter.VULN_CRIME,
                               Parameter.VULN_BREACH,
                               Parameter.VULN_POODLE,
                               Parameter.VULN_TLS_FALLBACK,
                               Parameter.VULN_SWEET32,
                               Parameter.VULN_FREAK,
                               Parameter.VULN_DROWN,
                               Parameter.VULN_LOGJAM,
                               Parameter.VULN_BEAST,
                               Parameter.VULN_LUCKY13,
                               Parameter.VULN_RC4,
                               *extra_parameters)
        scan_result = scanner.scan(target)
        findings = ScanResultFindings(scan_result, self.logger)

        vulns = {
            'BEAST': None,
            'BREACH': None,
            'CRIME': None,
            'DROWN': None,
            'FREAK': None,
            'Logjam': None, # ?
            'POODLE': None,
            'Sweet32': None,
            'TLS_FALLBACK_SCSV': None,
            'Lucky13': None,
            'RC4': None,
            'Renegotiation': None,
            'Renegotiation_client': None,
        }

        beast = findings.get('BEAST')
        if beast:
            if 'VULNERABLE' in beast:
                vulns['BEAST'] = {'vulnerable': True, 'ciphers': None}
                beast_ciphers = findings.get('BEAST_CBC_TLS1')
                if beast_ciphers:
                    vulns['BEAST']['ciphers'] = beast_ciphers.split()
            else:
                vulns['BEAST'] = {'vulnerable': False}

        vuln_no_extra = {
            'BREACH': 'BREACH',
            'CRIME': 'CRIME_TLS',
            'Sweet32': 'SWEET32',
            'DROWN': 'DROWN',
            'FREAK': 'FREAK',
            'POODLE': 'POODLE_SSL',  # TODO: handle POODLE_TLS once implemented
            'RC4': 'RC4',
            'Renegotiation': 'secure_renego',
            'Renegotiation_client': 'secure_client_renego'
        }
        for vuln_name, vuln_key in vuln_no_extra.items():
            finding = findings.get(vuln_key)
            if finding:
                vulns[vuln_name] = {'vulnerable': 'VULNERABLE' in finding}

        sweet32 = findings.get('SWEET32', ('uses 64 bit block ciphers',
                                           'not vulnerable'))
        if sweet32:
            vulns['Sweet32'] = {'vulnerable': sweet32 == 'uses 64 bit block ciphers'}

        logjam = findings.get('LOGJAM')
        if logjam:
            login_common_primes = findings.get('LOGJAM-common_primes')
            # TODO: Implement me

        fallback = findings.get('fallback_SCSV')
        if fallback:
            # Do we really want to classify this as an vulnerability like
            # testssl.sh does?
            vulns['TLS_FALLBACK_SCSV'] = {'vulnerable': fallback != 'supported' and
                                          'no protocol below ' not in fallback}

        lucky13 = findings.get('LUCKY13')
        if lucky13:
            # experimental; can be a false positive with a fixed openssl version
            vulns['Lucky13'] = {'vulnerable': 'potentially vulnerable' in lucky13}

        return vulns

    def _scan_stage_vulns_ids(self, target, extra_parameters):
        """Stage 2 scan: Contains vulnerabilities that could trigger an IDS"""
        scanner = TestsslshScanner(self._install_dir)
        scanner.add_parameters(Parameter.VULN_HEARTBLEED,
                               Parameter.VULN_CCS_INJECTION,
                               Parameter.VULN_TICKETBLEED,
                               Parameter.VULN_ROBOT,
                               *extra_parameters)
        scan_result = scanner.scan(target)
        findings = ScanResultFindings(scan_result, self.logger)

        vulns = {
            'Heartbleed': None,
            'CCS_Injection': None,
            'Ticketbleed': None,
            'ROBOT': None
        }
        vuln_no_extra = {
            'Heartbleed': 'heartbleed',
            'Ticketbleed': 'ticketbleed',
            'ROBOT': 'ROBOT'
        }
        for vuln_name, vuln_key in vuln_no_extra.items():
            finding = findings.get(vuln_key)
            if finding:
                vulns[vuln_name] = {'vulnerable': 'VULNERABLE' in finding}

        ccs = findings.get('CCS')
        if ccs:
            vulns['CCS_Injection'] = {'vulnerable': 'VULNERABLE' in ccs}
            if 'likely VULNERABLE with' in ccs:
                alert_byte = ccs.replace('likely VULNERABLE with ', '')
                vulns['CCS_Injection']['alert_type'] = 'unknown_alert_0x{}'.format(alert_byte)
            elif 'likely VULNERABLE' in ccs:
                vulns['CCS_Injection']['alert_type'] = 'bad_record_mac'

            if 'probably not vulnerable but' in ccs:
                byte_start = ccs.find('0x')
                byte_end = ccs.find(' ', byte_start)
                if byte_start > 0 and byte_end > 0:
                    alert_byte = ccs[byte_start:byte_end]
                    if alert_byte == '0x0A':
                        alert_type = 'unexpected_message'
                    elif alert_byte == '0x28':
                        alert_type = 'handshake_failure'
                    else:
                        alert_type = 'unknown_alert_{}'.format(alert_byte)
                    vulns['CCS_Injection']['alert_type'] = alert_type

        return vulns

    def update_dependencies(self):
        install_base_dir = self.options['install_base_dir']
        install_base_dir.mkdir(parents=True, exist_ok=True)
        hash_symlink = (install_base_dir / self.options['download_hash'])
        if hash_symlink.exists():
            self.logger.info('Expected testssl.sh version is already installed.')
            return
        self.logger.info('Downloading testssl.sh from %s', self.options['download_url'])
        with io.BytesIO() as f:
            download_file(self.options['download_url'], f,
                          verify_hash=self.options['download_hash'])
            f.seek(0)
            with tarfile.open(fileobj=f) as tarball:
                directory_name = Path(tarball.next().name).parts[0]
                tarball.extractall(path=install_base_dir)
        hash_symlink.symlink_to(directory_name, target_is_directory=True)
        self.logger.info('Successfully installed testssl.sh')

    def _get_host(self, result):
        raise NotImplemented


class ScanResultFindings:
    def __init__(self, scan_result, logger):
        self._scan_result = scan_result
        self._logger = logger

    def get(self, key, assertions=None):
        if key not in self._scan_result:
            return None
        finding = self._scan_result[key]['finding']
        if assertions is not None:
            if isinstance(assertions, (list, tuple)):
                if finding not in assertions:
                    msg = 'Finding for %s is not expected: %s' % (key, finding)
                    self._logger.error(msg)
                    raise ValueError(msg)
        return finding



import enum
import json
import subprocess
import tempfile
from pathlib import Path


class TestsslshFailed(Exception):
    def __init__(self, exit_code, *args):
        self.exit_code = exit_code
        super().__init__(*args)


class TestsslshFailedPartially(TestsslshFailed):
    def __init__(self, exit_code, partial_result, *args):
        self.partial_result = partial_result
        super().__init__(exit_code, *args)


class Parameter(enum.Enum):
    STARTTLS = '-t'
    PROTOCOLS = '-p'
    SERVER_DEFAULTS = '-S'
    SERVER_PREFERENCE = '-P'
    STANDARD_CIPHERS = '-s'
    CHECK_FORWARD_SECRECY = '-f'
    FAST = '--fast'
    IP = '--ip'
    SNEAKY = '--sneaky'
    # PHONE_OUT allows OCSP querying and CRL download
    PHONE_OUT = '--phone-out'
    JSONFILE = '--jsonfile'
    HTMLFILE = '--htmlfile'

    VULN_ALL = '-U'
    VULN_HEARTBLEED = '-H'
    VULN_CCS_INJECTION = '-I'
    VULN_TICKETBLEED = '-T'
    VULN_ROBOT = '-BB'
    VULN_CRIME = '-C'
    VULN_BREACH = '-B'
    VULN_POODLE = '-O'
    VULN_TLS_FALLBACK = '-Z'
    VULN_RENEGOTIATION = '-R'
    VULN_SWEET32 = '-W'
    VULN_BEAST = '-A'
    VULN_LUCKY13 = '-L'
    VULN_FREAK = '-F'
    VULN_LOGJAM = '-J'
    VULN_DROWN = '-D'
    VULN_RC4 = '-4'


class TestsslshScanner:
    def __init__(self, install_dir):
        self._install_dir = Path(install_dir)
        self.parameters = []
        self.environment = {
            'USLEEP_SND': '0.1', # 0.5
            'USLEEP_REC': '0.1',
            'TESTSSL_INSTALL_DIR': str(self._install_dir)
        }
        self.result = None

    def add_parameters(self, *parameters):
        for parameter in parameters:
            if isinstance(parameter, Parameter):
                parameter = parameter.value
            self.parameters.append(parameter)

    def scan(self, target_url):
        executable = self._install_dir / 'testssl.sh'
        command = [str(executable)] + self.parameters
        with tempfile.NamedTemporaryFile() as f:
            command += (Parameter.JSONFILE.value, f.name, target_url)
            p = subprocess.run(command,
                               stderr=subprocess.PIPE,
                               stdout=subprocess.DEVNULL,
                               env=self.environment,
                               check=False,
                               encoding='utf-8',
                               errors='replace')
            f.seek(0)
            try:
                scan_list = json.load(f)
            except json.JSONDecodeError:
                raise TestsslshFailed(-1000, 'JSON decode failed.')

        result = {}
        for entry in scan_list:
            if 'id' not in entry:
                continue
            result[entry['id']] = entry

        # Check if we actually have any results. This means that we check
        # whether there are keys that are not engine_problem or scanTime,
        # which are available even if the host is not reachable.
        min_length = 1
        for no_result_key in ('engine_problem', 'scanTime'):
            if no_result_key in result:
                min_length += 1
        if len(result) < min_length:
            result = None

        if not (0 <= p.returncode < 50):
            if result is not None:
                raise TestsslshFailedPartially(p.returncode, result, p.stderr)
            raise TestsslshFailed(p.returncode, p.stderr)

        return result



"""
Test for common server leaks.
"""
from concurrent.futures import ThreadPoolExecutor
from urllib.parse import urlparse

import requests
from requests.exceptions import ConnectionError
from requests.models import Response
from tldextract import extract

from privacyscanner.scanmodules import ScanModule


class ServerleaksScanModule(ScanModule):
    name = 'serverleak'
    dependencies = ['chromedevtools']
    required_keys = ['final_url']

    def scan_site(self, result, meta):
        scan_site(result, self.logger, self.options, meta)


def _match_db_dump(content):
    targets = ["SQLite", "CREATE TABLE", "INSERT INTO", "DROP TABLE"]
    matched = False
    for target in targets:
        matched |= target in content
    return matched


def _concat_sub(url, suffix):
    url_extract = extract(url)
    if url_extract.subdomain == "":
        return None
    site = url_extract.subdomain + "." + url_extract.domain
    return site + suffix


def _concat_full(url, suffix):
    url_extract = extract(url)
    site = url_extract.domain + "." + url_extract.suffix
    if url_extract.subdomain != "":
        site = url_extract.subdomain + "." + site
    return site + suffix


def _gen_db_domain_sql(url):
    return extract(url).domain + ".sql"


def _gen_db_sub_domain_sql(url):
    return _concat_sub(url, ".sql")


def _gen_db_full_domain_sql(url):
    return _concat_full(url, ".sql")


def _gen_db_domain_db(url):
    return extract(url).domain + ".db"


def _gen_db_sub_domain_db(url):
    return _concat_sub(url, ".db")


def _gen_db_full_domain_db(url):
    return _concat_full(url, ".db")


def _gen_db_domain_key(url):
    return extract(url).domain + ".key"


def _gen_db_sub_domain_key(url):
    return _concat_sub(url, ".key")


def _gen_db_full_domain_key(url):
    return _concat_full(url, ".key")


def _gen_db_domain_pem(url):
    return extract(url).domain + ".pem"


def _gen_db_sub_domain_pem(url):
    return _concat_sub(url, ".pem")


def _gen_db_full_domain_pem(url):
    return _concat_full(url, ".pem")


TRIALS = [
    ('server-status/', 'Apache Server Status'),
    ('server-info/', 'Apache Server Information'),
    ('test.php', 'phpinfo()'),
    ('phpinfo.php', 'phpinfo()'),
    ('.git/HEAD', 'ref:'),
    ('.svn/wc.db', 'SQLite'),
    ('core', 'ELF'),
    ('.DS_Store', 'Bud1'),

    # Check for Database dumps
    # sqldump - MySQL/MariaDB
    ('dump.db', _match_db_dump),
    ('dump.sql', _match_db_dump),
    ('sqldump.sql', _match_db_dump),
    ('sqldump.db', _match_db_dump),
    # SQLite
    ('db.sqlite', _match_db_dump),
    ('data.sqlite', _match_db_dump),
    ('sqlite.db', _match_db_dump),
    (_gen_db_domain_sql, _match_db_dump),
    (_gen_db_sub_domain_sql, _match_db_dump),
    (_gen_db_full_domain_sql, _match_db_dump),
    (_gen_db_domain_db, _match_db_dump),
    (_gen_db_sub_domain_db, _match_db_dump),
    (_gen_db_full_domain_db, _match_db_dump),

    # TODO PostgreSQL etc., additional common names

    # TLS Certs
    ('server.key', '-----BEGIN'),
    ('privatekey.key', '-----BEGIN'),
    ('private.key', '-----BEGIN'),
    ('myserver.key', '-----BEGIN'),
    ('key.pem', '-----BEGIN'),
    ('privkey.pem', '-----BEGIN'),
    (_gen_db_domain_key, '-----BEGIN'),
    (_gen_db_sub_domain_key, '-----BEGIN'),
    (_gen_db_full_domain_key, '-----BEGIN'),
    (_gen_db_domain_pem, '-----BEGIN'),
    (_gen_db_sub_domain_pem, '-----BEGIN'),
    (_gen_db_full_domain_pem, '-----BEGIN'),

    # Docker
    # https://infosec.rm-it.de/2018/08/19/scanning-the-alexa-top-1m-sites-for-dockerfiles/
    ('Dockerfile', 'FROM'),
    # https://twitter.com/svblxyz/status/1045013939904532482
    ('docker.env', '='),
    ('.env', '='),
    # Docker Compose
    ('docker-compose.yml', 'version:'),
]


def _get(url, timeout):
    try:
        response = requests.get(url, timeout=timeout)
        return response
    except ConnectionError:
        return None


def _response_to_json(resp: Response):
    """Generate a json byte string from a response
    received through requests."""
    # we store only the top of the file
    # because core dumps can become very large
    # also: we do not want to store more potentially sensitive data
    # than necessary to determine whether there is a leak or not

    return {
        'text': resp.content[0:50*1024].decode(errors='replace'),
        'status_code': resp.status_code,
        'headers': dict(resp.headers),
        'url': resp.url,
    }


def _check_leaks(url, max_workers):
    leaks = []
    trials = {}
    # determine hostname
    parsed_url = urlparse(url)

    with ThreadPoolExecutor(max_workers=max_workers) as executor:
        url_to_future = {}

        for trial, pattern in TRIALS:
            trial_t = trial
            # Check if trial is callable. If so, call it and save the result
            if callable(trial):
                trial_t = trial(url)
                if trial_t is None:
                    continue
            request_url = '{}://{}/{}'.format(
                parsed_url.scheme, parsed_url.netloc, trial_t)
            url_to_future[trial_t] = executor.submit(_get, request_url, 10)

        for trial in url_to_future:
            try:
                # response = requests.get(request_url, timeout=10)
                response = url_to_future[trial].result()
                if response is None:
                    continue

                match_url = '{}/{}'.format(parsed_url.netloc, trial)

                if match_url not in response.url:
                    # There has been a redirect.
                    continue

                trials[trial] = _response_to_json(response)
            except Exception:
                continue

    for trial, pattern in TRIALS:
        if callable(trial):
            trial = trial(url)
            if trial is None:
                continue
        if trial not in trials:
            # Test raw data too old or particular request failed.
            continue

        response = trials[trial]

        if response['status_code'] == 200:
            # The pattern can have three different types.
            # - If it is a simple string,
            #   we only check if it is contained in the response
            if isinstance(pattern, str):
                if pattern in response['text']:
                    leaks.append(trial)
            # - If it is callable,
            #   we call it with the response text and check the return value
            elif callable(pattern):
                if pattern(response['text']):
                    leaks.append(trial)

    return leaks

def scan_site(result, logger, options, meta):
    max_workers = options.get('max_workers', 8)

    # Note: This does not scan the original site_url before redirection.
    #       There might be cases where only the start page redirects, but
    #       other paths (which do not get redirected) contain sensitive files.
    result['leaks'] = _check_leaks(result['final_url'], max_workers)



from urllib.parse import urlparse

from privacyscanner.scanmodules import ScanModule


class ExampleScanModule(ScanModule):
    name = 'example'
    dependencies = []
    required_keys = ['site_url']

    def scan_site(self, result, meta):
        """Scans a site and adds more information to the result.

        The parameter result behaves like a dictionary, you can set keys on it
        and call the usual methods on dicts. If you change a non-shallow key,
        you have to mark the underlying shallow key as dirty by calling
        result.mark_dirty(shallow_key).

        Furthermore result exposes a logger on the logger parameter where you can
        send log messages to the scanning master.

        For storing files, you can call result.add_file(identifier, filecontents)
        which will send filecontents to the master. If you provide a file-like
        object it will read from the file and send the file contents to the
        master. The identifier represents a file name and must be unique within
        a scan. To store files for debug purposes, call result.add_debug_file
        instead, which has the same API.

        To start with, you can access result['site_url'], which is populated by the
        master.

        The parameter options is a dictionary with the configuration of the scan
        module as specified in the configuration file at SCAN_MODULE_OPTIONS.
        """
        self.logger.info('we will check the site for https')
        parsed_site = urlparse(result['site_url'])
        result['is_https'] = parsed_site.scheme == 'https'
        # result.add_file('screenshot.png')
        if self.options.get('save_nops'):
            result.add_debug_file('nops.bin', b'\x90\x90\x90\x90')



import warnings
from urllib.parse import urlparse

import pychrome

from privacyscanner.scanmodules.chromedevtools.extractors.base import Extractor
from privacyscanner.scanmodules.chromedevtools.utils import scripts_disabled


ELEMENT_NODE = 1


class ImprintExtractor(Extractor):
    IMPRINT_KEYWORDS = ['imprint', 'impressum', 'contact', 'kontakt', 'about us', 'über uns']

    def extract_information(self):
        # Disable scripts to avoid DOM changes while searching for the imprint.
        with scripts_disabled(self.page.tab, self.options):
            self._extract_imprint()

    def _extract_imprint(self):
        node_id = self.page.tab.DOM.getDocument()['root']['nodeId']
        links = self.page.tab.DOM.querySelectorAll(nodeId=node_id, selector='a')['nodeIds']
        imprint_link = None

        # Use the browsers search to search for the keywords. For each result,
        # we walk up the DOM until we find an ``a'' element. If this element
        # has an href, this is our imprint link. Otherwise, we look for the
        # next search result.
        for keyword in self.IMPRINT_KEYWORDS:
            search = self.page.tab.DOM.performSearch(query=keyword)
            if search['resultCount'] == 0:
                continue
            results = self.page.tab.DOM.getSearchResults(
                searchId=search['searchId'], fromIndex=0, toIndex=search['resultCount'])
            for node_id in results['nodeIds']:
                while node_id is not None:
                    try:
                        node = self.page.tab.DOM.describeNode(nodeId=node_id)['node']
                    except pychrome.CallMethodException:
                        # For some reason, nodes seem to disappear in-between,
                        # so just ignore these cases.
                        break
                    if node['nodeType'] == ELEMENT_NODE and node['nodeName'].lower() == 'a':
                        if not self._is_visible(node_id):
                            break
                        href = self._get_href(node_id)
                        if href:
                            imprint_link = href
                        break
                    node_id = node.get('parentId')
                if imprint_link:
                    break
            if imprint_link:
                break

        # If our browser search does not give results, search more brutally
        # for all links, including those, who are not visible to the user.
        if not imprint_link:
            for link in links:
                try:
                    link_html = self.page.tab.DOM.getOuterHTML(nodeId=link)['outerHTML']
                except pychrome.CallMethodException:
                    # For some reason, nodes seem to disappear in-between,
                    # so just ignore these cases.
                    break
                for order_id, keyword in enumerate(self.IMPRINT_KEYWORDS):
                    if keyword in link_html:
                        href = self._get_href(link)
                        if href:
                            imprint_link = href
                            break
                if imprint_link:
                    break

        if imprint_link:
            if imprint_link.startswith('//'):
                p = urlparse(self.result['final_url'])
                imprint_link = '{}:{}'.format(p.scheme, imprint_link)
            elif imprint_link.startswith('/'):
                p = urlparse(self.result['final_url'])
                imprint_link = '{}://{}{}'.format(p.scheme, p.hostname, imprint_link)
            elif imprint_link.startswith(('https://', 'http://')):
                # Nothing to do, already the full URL
                pass
            else:
                base_url = self.result['final_url'].rsplit('/', 1)[0]
                imprint_link = '{}/{}'.format(base_url, imprint_link)
        self.result['imprint_url'] = imprint_link

    def _get_href(self, node_id):
        attrs = self.page.tab.DOM.getAttributes(nodeId=node_id)['attributes']
        attrs = dict(zip(*[iter(attrs)]*2))
        return attrs.get('href')

    def _is_visible(self, node_id):
        try:
            with warnings.catch_warnings():
                warnings.simplefilter("ignore")
                self.page.tab.DOM.getBoxModel(nodeId=node_id)
            return True
        except pychrome.exceptions.CallMethodException:
            return False



import ssl
from collections import namedtuple
import smtplib
from pathlib import Path

from privacyscanner.scanmodules import ScanModule
from privacyscanner.utils import set_default_options
from privacyscanner.utils.tls import get_cipher_info, get_certificate_info


LINUX_CA_FILE = Path('/etc/ssl/certs/ca-certificates.crt')

MailserverResult = namedtuple('MailserverResult',
                              ['banner', 'cipher', 'certificate', 'features',
                               'error'])

class MailScanModule(ScanModule):
    name = 'mail'
    required_keys = ['mail', 'dns']
    dependencies = ['dns']

    def __init__(self, options):
        ca_file = None
        if LINUX_CA_FILE.exists():
            ca_file = str(LINUX_CA_FILE)
        set_default_options(options, {
            'local_hostname': None,
            'timeout': 10,
            'ca_file': ca_file,
            'ca_path': None
        })
        super().__init__(options)

    def scan_site(self, result, meta):
        mail = result['mail']
        try:
            # MX records are ordered by priority (most preferred first)
            mail_host = result['dns'][mail['domain']]['MX'][0]['host']
        except (KeyError, IndexError):
            # We have either an error when receiving MX records
            # or have no MX records.
            mail_host = mail['domain']

        has_cas = (self.options['ca_file'] is not None or
                   self.options['ca_path'] is not None)
        if not has_cas:
            self.logger.warning('No CA certificates loaded. Cannot check for trust.')

        conn = smtplib.SMTP(local_hostname=self.options['local_hostname'],
                            timeout=self.options['timeout'])
        try:
            # The Python API of smtplib has a flaw: If you pass the host
            # to the initializer, it will connect and set self._host to
            # the host, but will not return the first message after connecting.
            # We want that message, because it might contain a version number.
            # You get the first message if you do not pass the host to the
            # initializer, but call connect with the host yourself. However,
            # this does not set self._host and therefore .starttls() fails.
            # So we add this ugly hack, which sets self._host before connect.
            conn._host = mail_host
            code, banner = conn.connect(mail_host)
            mail['banner'] = banner.decode('utf-8', errors='replace')
            conn.ehlo_or_helo_if_needed()
            has_starttls = conn.has_extn('STARTTLS')
            mail['has_starttls'] = has_starttls
            if has_starttls:
                context = ssl.create_default_context(
                    cafile=self.options['ca_file'],
                    capath=self.options['ca_path']
                )
                context.check_hostname = has_cas
                context.verify_mode = ssl.CERT_REQUIRED if has_cas else ssl.CERT_NONE
                try:
                    conn.starttls(context=context)
                    is_trusted = True if has_cas else None
                except ssl.CertificateError:
                    is_trusted = False
                    context.check_hostname = False
                    context.verify_mode = ssl.CERT_NONE
                    conn.connect(mail_host)
                    conn.starttls(context=context)
                conn.ehlo_or_helo_if_needed()

                mail.update(get_cipher_info(conn.sock.cipher()))
                cert_der = conn.sock.getpeercert(binary_form=True)
                mail['certificate'] = get_certificate_info(cert_der)
                mail['certificate']['is_trusted'] = is_trusted
            code, msg = conn.verify('root')
            mail['allows_vrfy'] = code in (250, 251, 252, 550, 551, 553)
            code, msg = conn.expn('admin')
            mail['allows_expn'] = code in (250, 550)
        except smtplib.SMTPHeloError:
            mail['error'] = 'EHLO'
        except smtplib.SMTPException:
            mail['error'] = 'other'
        finally:
            conn.close()

        result.mark_dirty('mail')



import tarfile
import tempfile
from pathlib import Path

from dns import resolver, reversename
from dns.exception import DNSException
from geoip2.database import Reader
from geoip2.errors import AddressNotFoundError

from privacyscanner.scanmodules import ScanModule
from privacyscanner.scanmodules.chromedevtools import parse_domain, TLDEXTRACT_CACHE_FILE
from privacyscanner.utils import set_default_options, copy_to, download_file, file_is_outdated


GEOIP_DATABASE_PATH = Path('GeoIP/GeoLite2-Country.mmdb')
GEOIP_DOWNLOAD_URL = 'https://geolite.maxmind.com/download/geoip/database/GeoLite2-Country.tar.gz'
GEOIP_MAX_AGE = 3 * 24 * 3600


class DNSScanModule(ScanModule):
    name = 'dns'
    dependencies = ['chromedevtools']
    required_keys = ['site_url', 'redirect_chain']

    def __init__(self, options):
        set_default_options(options, {
            'geoip_download_url': GEOIP_DOWNLOAD_URL,
            'geoip_max_age': GEOIP_MAX_AGE,
        })
        super().__init__(options)
        cache_file = self.options['storage_path'] / TLDEXTRACT_CACHE_FILE
        parse_domain.cache_file = str(cache_file)
        try:
            geoip_path = Path(options['geoip_database_path'])
        except KeyError:
            geoip_path = self.options['storage_path'] / GEOIP_DATABASE_PATH
        self.options['geoip_database_path'] = geoip_path
        self._geoip_reader = None

    def scan_site(self, result, meta):
        dns = {}

        # Fetch MX records
        p = parse_domain(result['site_url'])
        mail_domain = p.fqdn[len('www.'):] if p.fqdn.startswith('www.') else p.fqdn
        mx_records = self._get_mx_records(mail_domain)

        # Create a list for which we fetch A/AAAA records
        domain_list = {mail_domain}
        domain_list.update(parse_domain(url).fqdn for url in result['redirect_chain'])
        if mx_records is not None:
            domain_list.update(record['host'] for record in mx_records)

        # Fetch A/AAAA records and reverse (PTR)
        for url in domain_list:
            p = parse_domain(url)
            records = dns.setdefault(p.fqdn, {})
            records['A'] = self._get_dns_records(p.fqdn, 'A')
            records['AAAA'] = self._get_dns_records(p.fqdn, 'AAAA')

        dns.setdefault(mail_domain, {})['MX'] = mx_records

        result['mail'] = {'domain': mail_domain}
        result['dns'] = dns

    def update_dependencies(self):
        geoip_database_path = self.options['geoip_database_path']
        geoip_max_age = self.options['geoip_max_age']
        if not file_is_outdated(geoip_database_path, geoip_max_age):
            return
        geoip_database_path.parent.mkdir(parents=True, exist_ok=True)
        FILES = ['COPYRIGHT.txt', 'LICENSE.txt', 'GeoLite2-Country.mmdb']
        with tempfile.NamedTemporaryFile() as f:
            download_file(GEOIP_DOWNLOAD_URL, f)
            archive = tarfile.open(f.name)
            for member in archive.getmembers():
                base_name = Path(member.name).name
                if base_name in FILES and member.isfile():
                    with (geoip_database_path.parent / base_name).open('wb') as f:
                        copy_to(archive.extractfile(member), f)

    def _get_geoip_reader(self):
        if self._geoip_reader is None:
            self._geoip_reader = Reader(str(self.options['geoip_database_path']))
        return self._geoip_reader

    def _get_dns_records(self, qname, rdtype):
        reader = self._get_geoip_reader()
        try:
            answer = resolver.query(qname, rdtype)
        except (resolver.NXDOMAIN, resolver.NoAnswer):
            return []
        except DNSException as e:
            self.logger.exception('Could not get %(rdtype) records for %(qname)s: %(msg)s',
                                  {'qname': qname, 'rdtype': rdtype, 'msg': str(e)})
            return None
        entries = []
        for a in answer:
            country = None
            continent = None
            try:
                geo_result = reader.country(a.address)
                country = geo_result.country.name
                continent = geo_result.continent.name
            except AddressNotFoundError:
                pass
            entries.append({
                'ip': a.address,
                'reverse': self._get_reverse_records(a.address),
                'country': country,
                'continent': continent
            })
        return entries

    def _get_reverse_records(self, address):
        qname = reversename.from_address(address)
        try:
            answer = resolver.query(qname, 'PTR')
        except (resolver.NXDOMAIN, resolver.NoAnswer):
            return []
        except DNSException as e:
            self.logger.exception('Could not get PTR records for %s: %s', address, str(e))
            return None
        return [a.target.to_text()[:-1] for a in answer]

    def _get_mx_records(self, mail_domain):
        try:
            answer = resolver.query(mail_domain, 'MX')
        except (resolver.NXDOMAIN, resolver.NoAnswer):
            return []
        except DNSException as e:
            self.logger.exception('Could not get MX records for %s: %s', mail_domain, str(e))
            return None
        mx_records = [{
            'priority': a.preference,
            'host': a.exchange.to_text()[:-1]
        } for a in answer]
        # We include the name in the ordering to have a deterministic order
        mx_records.sort(key=lambda rec: (rec['priority'], rec['host']))
        return mx_records



from privacyscanner.scanmodules.chromedevtools.utils import parse_domain
from privacyscanner.scanmodules.chromedevtools.extractors.base import Extractor


class ThirdPartyExtractor(Extractor):
    def extract_information(self):
        third_parties = {
            'fqdns': set(),
            'num_http_requests': 0,
            'num_https_requests': 0
        }
        first_party_domains = set()
        for url in (self.result['site_url'], self.result['final_url']):
            extracted = parse_domain(url)
            first_party_domains.add(extracted.registered_domain)
        for request in self.page.request_log:
            request['is_thirdparty'] = False
            extracted_url = parse_domain(request['url'])
            parsed_url = request['parsed_url']
            if extracted_url.registered_domain in first_party_domains:
                continue
            if request['url'].startswith('data:'):
                continue
            request['is_thirdparty'] = True
            third_parties['fqdns'].add(extracted_url.fqdn)
            if parsed_url.scheme not in ('http', 'https'):
                continue
            third_parties['num_{}_requests'.format(parsed_url.scheme)] += 1
        third_parties['fqdns'] = list(third_parties['fqdns'])
        third_parties['fqdns'].sort()
        self.result['third_parties'] = third_parties

        for cookie in self.result['cookies']:
            domain = cookie['domain']
            if domain.startswith('.'):
                domain = domain[1:]
            domain = parse_domain(domain).registered_domain
            cookie['is_thirdparty'] = domain not in first_party_domains



from privacyscanner.scanmodules.chromedevtools.extractors.base import Extractor
from privacyscanner.scanmodules.chromedevtools.utils import camelcase_to_underscore


class TLSDetailsExtractor(Extractor):
    def extract_information(self):
        redirects_secure = None
        redirects_insecure = None
        if self.result['site_url'].startswith('http://'):
            redirects_secure = self.result['final_url'].startswith('https://')
            if redirects_secure:
                redirects_insecure = False
        if self.result['site_url'].startswith('https://'):
            redirects_insecure = self.result['final_url'].startswith('http://')
            if redirects_insecure:
                redirects_secure = False

        self.result['https'] = {
            'has_tls': None,
            'redirects_secure': redirects_secure,
            'redirects_insecure': redirects_insecure
        }

        if redirects_insecure:
            response = self.page.response_log[0]
        else:
            response = self.page.final_response
        if response is None:
            self.logger.error('Could not get response for final_url')
            return

        if 'securityDetails' not in response:
            self.result['https']['has_tls'] = False
            return

        details = {}
        # See https://chromedevtools.github.io/devtools-protocol/tot/Network#type-SecurityDetails
        properties = [
            # Protocol name (e.g. "TLS 1.2" or "QUIC")
            'protocol',
            # Key Exchange used by the connection, or the empty string
            # if not applicable.
            'keyExchange',
            # (EC)DH group used by the connection, if applicable.
            'keyExchangeGroup',
            # Cipher name.
            'cipher',
            # TLS MAC. Note that AEAD ciphers do not have separate MACs.
            'mac',
            # TODO: The following might be also interesting
            # signedCertificateTimestampList
            # certificateTransparencyCompliance
        ]

        for key, value in response['securityDetails'].items():
            if key not in properties:
                continue
            details[camelcase_to_underscore(key)] = value
        if 'mac' not in details:
            details['mac'] = None

        self.result['https']['has_tls'] = True
        self.result['https'].update(details)



from datetime import timedelta, datetime


class RetryScan(Exception):
    """Raise this exception in scan modules to run the scanning module again.

    Note: This will not retry a scan in case the maximum number of (re)tries
    (default: 3) is already reached.
    """
    pass


class RescheduleLater(Exception):
    def __init__(self, not_before, *args):
        super().__init__(*args)
        if isinstance(not_before, int):
            not_before = timedelta(seconds=not_before)
        if isinstance(not_before, timedelta):
            not_before = datetime.utcnow() + not_before
        self.not_before = not_before



import hashlib
import os
import errno
import fcntl
import re
import time
from base64 import b32encode
from urllib.request import Request, urlopen


FAKE_UA = 'Mozilla/5.0 (X11; Linux x86_64; rv:61.0) Gecko/20100101 Firefox/61.0'


class DownloadVerificationFailed(Exception):
    pass


class NumericLock:
    def __init__(self, lock_dir):
        self.lock_dir = lock_dir
        self._lock_file = None

    def __enter__(self):
        i = 0
        while True:
            i += 1
            try:
                f = open(self.lock_dir / ('%d.lock' % i), 'wb')
                fcntl.lockf(f.fileno(), fcntl.LOCK_EX | fcntl.LOCK_NB)
                self._lock_file = f
                return i
            except OSError as e:
                if e.errno in (errno.EACCES, errno.EAGAIN):
                    continue
                raise

    def __exit__(self, exc_type, exc_val, exc_tb):
        fcntl.lockf(self._lock_file.fileno(), fcntl.LOCK_UN)
        self._lock_file.close()
        os.unlink(self._lock_file.name)


def download_file(url, fileobj, verify_hash=None):
    request = Request(url)
    request.add_header('User-Agent', FAKE_UA)
    response = urlopen(request)
    hasher = hashlib.sha256() if verify_hash else None
    copy_to(response, fileobj, hasher)
    fileobj.flush()
    if verify_hash and hasher.hexdigest() != verify_hash:
        msg = 'SHA256({!r}) does not match {}.'.format(url, verify_hash)
        raise DownloadVerificationFailed(msg)


def copy_to(src, dest, hasher=None):
    while True:
        data = src.read(8192)
        if not data:
            break
        dest.write(data)
        if hasher:
            hasher.update(data)


def file_is_outdated(path, max_age):
    try:
        return path.stat().st_mtime + max_age < time.time()
    except FileNotFoundError:
        return True


def set_default_options(target, defaults):
    for key, value in defaults.items():
        if key in target:
            new_target = target[key]
            if isinstance(new_target, dict):
                set_default_options(new_target, value)
        else:
            target[key] = value


def rand_str(length):
    rand_bits = os.urandom((length * 5 // 8) + 1)
    return b32encode(rand_bits).decode()[:length].lower()


def calculate_jaccard_index(a: bytes, b: bytes) -> float:
    """Calculate the jaccard similarity of a and b."""
    pattern = re.compile(rb'[ \n]')
    # remove tokens containing / to prevent wrong classifications for
    # absolute paths
    a = {token for token in pattern.split(a) if b'/' not in token}
    b = {token for token in pattern.split(b) if b'/' not in token}
    intersection = a.intersection(b)
    union = a.union(b)
    return len(intersection) / len(union)



from pathlib import Path

from adblockeval import AdblockRules

from privacyscanner.scanmodules.chromedevtools.utils import parse_domain
from privacyscanner.scanmodules.chromedevtools.extractors.base import Extractor
from privacyscanner.utils import download_file


EASYLIST_DOWNLOAD_PREFIX = 'https://easylist.to/easylist/'
EASYLIST_FILES = ['easylist.txt', 'easyprivacy.txt', 'fanboy-annoyance.txt']
EASYLIST_PATH = Path('easylist')

_adblock_rules_cache = None


class TrackerDetectExtractor(Extractor):
    def extract_information(self):
        self._load_rules()
        trackers_fqdn = set()
        trackers_domain = set()
        num_tracker_requests = 0
        blacklist = set()
        num_evaluations = 0
        for request in self.page.request_log:
            request['is_tracker'] = False
            if not request['is_thirdparty'] or request['url'].startswith('data:'):
                continue
            is_tracker = request['parsed_url'].netloc in blacklist
            if not is_tracker:
                # Giving only the first 150 characters of an URL is
                # sufficient to get good matches, so this will speed
                # up checking quite a bit!
                match_result = self.rules.match(request['url'][:150],
                                                request['document_url'])
                is_tracker = match_result.is_match
                num_evaluations += 1
            if is_tracker:
                request['is_tracker'] = True
                extracted = parse_domain(request['url'])
                if extracted.fqdn:
                    trackers_fqdn.add(extracted.fqdn)
                trackers_domain.add(extracted.registered_domain)
                num_tracker_requests += 1
                blacklist.add(request['parsed_url'].netloc)

        num_tracker_cookies = 0
        for cookie in self.result['cookies']:
            is_tracker = False
            domain = cookie['domain']
            if domain in trackers_fqdn or domain in trackers_domain:
                is_tracker = True
            elif domain.startswith('.'):
                reg_domain = parse_domain(domain[1:]).registered_domain
                if reg_domain in trackers_domain:
                    is_tracker = True

            if is_tracker:
                num_tracker_cookies += 1
            cookie['is_tracker'] = is_tracker

        self.result['tracking'] = {
            'trackers': list(sorted(trackers_fqdn)),
            'num_tracker_requests': num_tracker_requests,
            'num_tracker_cookies': num_tracker_cookies
        }

    def _load_rules(self):
        global _adblock_rules_cache

        if _adblock_rules_cache is not None:
            self.rules = _adblock_rules_cache
            return

        easylist_path = self.options['storage_path'] / EASYLIST_PATH
        easylist_files = [easylist_path / filename for filename in EASYLIST_FILES]
        self.rules = AdblockRules(rule_files=easylist_files,
                                  cache_file=easylist_path / 'rules.cache',
                                  skip_parsing_errors=True)
        _adblock_rules_cache = self.rules

    @staticmethod
    def update_dependencies(options):
        easylist_path = options['storage_path'] / EASYLIST_PATH
        easylist_path.mkdir(parents=True, exist_ok=True)
        for filename in EASYLIST_FILES:
            download_url = EASYLIST_DOWNLOAD_PREFIX + filename
            target_file = (easylist_path / filename).open('wb')
            download_file(download_url, target_file)
